"""
AIPTX Terminal Animations
=========================

Beautiful spinner animations, progress bars, and visual effects for CLI.

Usage:
    from aipt_v2.ui import Spinner, ProgressBar, MultiProgress, Colors

    # Spinner
    with Spinner("Scanning target...", style='cyber') as spinner:
        do_work()
        spinner.update("Found vulnerabilities...")

    # Progress Bar
    bar = ProgressBar(total=100, style='hacker')
    for i in range(101):
        bar.update(i, f"Processing {i}%")
    bar.finish()

    # Multi-task Progress
    mp = MultiProgress([
        {'name': 'Nmap', 'total': 100},
        {'name': 'Nikto', 'total': 100},
    ])
    mp.update(0, 50)
    mp.complete(1)
"""

from __future__ import annotations

import sys
import time
import threading
import random
from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any, Callable
from datetime import datetime


class Colors:
    """ANSI color codes for terminal styling."""

    # Reset
    RESET = "\033[0m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    REVERSE = "\033[7m"
    HIDDEN = "\033[8m"

    # Regular Colors
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    # Bright Colors
    BRIGHT_BLACK = "\033[90m"
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"

    # Background Colors
    BG_BLACK = "\033[40m"
    BG_RED = "\033[41m"
    BG_GREEN = "\033[42m"
    BG_YELLOW = "\033[43m"
    BG_BLUE = "\033[44m"
    BG_MAGENTA = "\033[45m"
    BG_CYAN = "\033[46m"
    BG_WHITE = "\033[47m"

    # Cursor Control
    HIDE_CURSOR = "\033[?25l"
    SHOW_CURSOR = "\033[?25h"
    CLEAR_LINE = "\033[2K"
    MOVE_UP = "\033[1A"
    MOVE_DOWN = "\033[1B"
    SAVE_POS = "\033[s"
    RESTORE_POS = "\033[u"

    # AIPTX Brand Colors
    AIPTX_CYAN = "\033[38;5;51m"
    AIPTX_PURPLE = "\033[38;5;135m"
    AIPTX_ORANGE = "\033[38;5;208m"
    AIPTX_PINK = "\033[38;5;198m"
    AIPTX_GREEN = "\033[38;5;46m"

    @classmethod
    def rgb(cls, r: int, g: int, b: int) -> str:
        """Create RGB color code."""
        return f"\033[38;2;{r};{g};{b}m"

    @classmethod
    def bg_rgb(cls, r: int, g: int, b: int) -> str:
        """Create RGB background color code."""
        return f"\033[48;2;{r};{g};{b}m"

    @classmethod
    def gradient(cls, text: str, start_color: tuple, end_color: tuple) -> str:
        """Apply gradient color to text."""
        result = []
        length = len(text)
        for i, char in enumerate(text):
            ratio = i / max(length - 1, 1)
            r = int(start_color[0] + (end_color[0] - start_color[0]) * ratio)
            g = int(start_color[1] + (end_color[1] - start_color[1]) * ratio)
            b = int(start_color[2] + (end_color[2] - start_color[2]) * ratio)
            result.append(f"\033[38;2;{r};{g};{b}m{char}")
        return "".join(result) + cls.RESET


# Spinner animation frames
SPINNER_STYLES = {
    # Braille dots
    "dots": ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"],
    "dots2": ["⣾", "⣽", "⣻", "⢿", "⡿", "⣟", "⣯", "⣷"],
    "dots3": ["⠁", "⠂", "⠄", "⡀", "⢀", "⠠", "⠐", "⠈"],
    "dots4": ["⠃", "⠊", "⠒", "⠢", "⠆", "⠰", "⠤", "⠖", "⠲", "⠴", "⠦", "⠧"],

    # Lines and arrows
    "line": ["-", "\\", "|", "/"],
    "line2": ["◐", "◓", "◑", "◒"],
    "arrow": ["←", "↖", "↑", "↗", "→", "↘", "↓", "↙"],
    "arrow2": ["⬆️ ", "↗️ ", "➡️ ", "↘️ ", "⬇️ ", "↙️ ", "⬅️ ", "↖️ "],

    # Shapes
    "circle": ["◜", "◠", "◝", "◞", "◡", "◟"],
    "square": ["◰", "◳", "◲", "◱"],
    "triangle": ["◢", "◣", "◤", "◥"],
    "star": ["✶", "✸", "✹", "✺", "✹", "✷"],

    # Blocks
    "block": ["▁", "▂", "▃", "▄", "▅", "▆", "▇", "█", "▇", "▆", "▅", "▄", "▃", "▁"],
    "block2": ["▖", "▘", "▝", "▗"],
    "pulse": ["█", "▓", "▒", "░", "▒", "▓"],
    "grow": ["▁", "▂", "▃", "▄", "▅", "▆", "▇", "█", "▇", "▆", "▅", "▄", "▃", "▂"],

    # Cyber/Hacker style
    "cyber": ["◢", "◣", "◤", "◥", "▪", "▫"],
    "hacker": ["⟨⟩", "⟪⟫", "⟬⟭", "⟮⟯"],
    "matrix": ["0", "1", "0", "1", "▓", "░"],
    "snake": ["⠏", "⠛", "⠹", "⢸", "⣰", "⣤", "⣆", "⡇"],

    # Bouncing
    "bounce": ["⠁", "⠂", "⠄", "⠂"],
    "bounce2": [".", "o", "O", "o"],

    # Clock
    "clock": ["🕐", "🕑", "🕒", "🕓", "🕔", "🕕", "🕖", "🕗", "🕘", "🕙", "🕚", "🕛"],

    # Simple
    "simple": ["◜ ", " ◝", " ◞", "◟ "],
}


class Spinner:
    """
    Animated spinner for terminal.

    Usage:
        # As context manager
        with Spinner("Loading...", style='cyber') as spinner:
            do_work()

        # Manual control
        spinner = Spinner("Scanning...")
        spinner.start()
        do_work()
        spinner.stop(success=True)
    """

    def __init__(
        self,
        message: str = "Loading...",
        style: str = "dots",
        color: str = Colors.BRIGHT_CYAN,
        speed: float = 0.1,
        success_symbol: str = "✓",
        fail_symbol: str = "✗",
        success_color: str = Colors.BRIGHT_GREEN,
        fail_color: str = Colors.BRIGHT_RED,
    ):
        self.message = message
        self.frames = SPINNER_STYLES.get(style, SPINNER_STYLES["dots"])
        self.color = color
        self.speed = speed
        self.success_symbol = success_symbol
        self.fail_symbol = fail_symbol
        self.success_color = success_color
        self.fail_color = fail_color

        self._running = False
        self._thread: Optional[threading.Thread] = None
        self._frame_index = 0
        self._start_time: Optional[float] = None

    def start(self) -> "Spinner":
        """Start the spinner animation."""
        if self._running:
            return self

        self._running = True
        self._start_time = time.time()
        sys.stdout.write(Colors.HIDE_CURSOR)
        sys.stdout.flush()

        self._thread = threading.Thread(target=self._spin, daemon=True)
        self._thread.start()
        return self

    def _spin(self):
        """Spinner animation loop."""
        while self._running:
            frame = self.frames[self._frame_index % len(self.frames)]
            elapsed = time.time() - self._start_time if self._start_time else 0
            elapsed_str = f" ({elapsed:.1f}s)" if elapsed > 1 else ""

            line = f"\r{self.color}{frame}{Colors.RESET} {self.message}{Colors.DIM}{elapsed_str}{Colors.RESET}"
            sys.stdout.write(Colors.CLEAR_LINE + line)
            sys.stdout.flush()

            self._frame_index += 1
            time.sleep(self.speed)

    def update(self, message: str):
        """Update spinner message."""
        self.message = message

    def stop(self, success: bool = True, final_message: str = None):
        """Stop the spinner with success/failure indicator."""
        self._running = False
        if self._thread:
            self._thread.join(timeout=0.5)

        elapsed = time.time() - self._start_time if self._start_time else 0
        elapsed_str = f" ({elapsed:.1f}s)"

        symbol = self.success_symbol if success else self.fail_symbol
        color = self.success_color if success else self.fail_color
        msg = final_message if final_message else self.message

        line = f"\r{color}{symbol}{Colors.RESET} {msg}{Colors.DIM}{elapsed_str}{Colors.RESET}\n"
        sys.stdout.write(Colors.CLEAR_LINE + line)
        sys.stdout.write(Colors.SHOW_CURSOR)
        sys.stdout.flush()

    def __enter__(self) -> "Spinner":
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop(success=exc_type is None)


# Progress bar styles
PROGRESS_STYLES = {
    "cyber": {
        "left": "【",
        "right": "】",
        "fill": "▰",
        "empty": "▱",
        "fill_color": Colors.BRIGHT_CYAN,
        "empty_color": Colors.DIM,
    },
    "hacker": {
        "left": "⟨",
        "right": "⟩",
        "fill": "█",
        "empty": "░",
        "fill_color": Colors.BRIGHT_GREEN,
        "empty_color": Colors.DIM,
    },
    "modern": {
        "left": "",
        "right": "",
        "fill": "━",
        "empty": "─",
        "fill_color": Colors.BRIGHT_MAGENTA,
        "empty_color": Colors.DIM,
    },
    "blocks": {
        "left": "",
        "right": "",
        "fill": "▓",
        "empty": "░",
        "fill_color": Colors.BRIGHT_YELLOW,
        "empty_color": Colors.DIM,
    },
    "minimal": {
        "left": "[",
        "right": "]",
        "fill": "=",
        "empty": " ",
        "fill_color": Colors.WHITE,
        "empty_color": Colors.DIM,
    },
    "dots": {
        "left": "⟨",
        "right": "⟩",
        "fill": "●",
        "empty": "○",
        "fill_color": Colors.BRIGHT_BLUE,
        "empty_color": Colors.DIM,
    },
    "arrows": {
        "left": "",
        "right": "▶",
        "fill": "═",
        "empty": "─",
        "fill_color": Colors.BRIGHT_CYAN,
        "empty_color": Colors.DIM,
    },
    "pulse": {
        "left": "⟦",
        "right": "⟧",
        "fill": "▮",
        "empty": "▯",
        "fill_color": Colors.AIPTX_CYAN,
        "empty_color": Colors.DIM,
    },
    "neon": {
        "left": "◄",
        "right": "►",
        "fill": "■",
        "empty": "□",
        "fill_color": Colors.AIPTX_PINK,
        "empty_color": Colors.DIM,
    },
    "fire": {
        "left": "🔥",
        "right": "",
        "fill": "█",
        "empty": "░",
        "fill_color": Colors.BRIGHT_RED,
        "empty_color": Colors.DIM,
    },
    "gradient": {
        "left": "",
        "right": "",
        "fill": "█",
        "empty": "░",
        "fill_color": None,  # Special handling
        "empty_color": Colors.DIM,
    },
}


class ProgressBar:
    """
    Beautiful progress bar with multiple styles.

    Usage:
        bar = ProgressBar(total=100, style='cyber')
        for i in range(101):
            bar.update(i, f"Processing {i}%")
            time.sleep(0.05)
        bar.finish("Complete!")
    """

    def __init__(
        self,
        total: int = 100,
        width: int = 30,
        style: str = "cyber",
        show_percentage: bool = True,
        show_eta: bool = True,
        show_count: bool = False,
        prefix: str = "",
        color: str = None,
    ):
        self.total = total
        self.width = width
        self.style_config = PROGRESS_STYLES.get(style, PROGRESS_STYLES["cyber"])
        self.show_percentage = show_percentage
        self.show_eta = show_eta
        self.show_count = show_count
        self.prefix = prefix
        self.color = color or self.style_config.get("fill_color", Colors.BRIGHT_CYAN)

        self.current = 0
        self._start_time = time.time()
        self._hidden_cursor = False

    def _format_time(self, seconds: float) -> str:
        """Format seconds into readable time."""
        if seconds < 60:
            return f"{seconds:.0f}s"
        elif seconds < 3600:
            return f"{seconds // 60:.0f}m {seconds % 60:.0f}s"
        else:
            return f"{seconds // 3600:.0f}h {(seconds % 3600) // 60:.0f}m"

    def _get_gradient_char(self, position: float) -> str:
        """Get gradient colored character."""
        # Red -> Yellow -> Green gradient
        if position < 0.5:
            r = 255
            g = int(255 * (position * 2))
            b = 0
        else:
            r = int(255 * (1 - (position - 0.5) * 2))
            g = 255
            b = 0
        return f"\033[38;2;{r};{g};{b}m{self.style_config['fill']}"

    def update(self, current: int, message: str = ""):
        """Update progress bar."""
        if not self._hidden_cursor:
            sys.stdout.write(Colors.HIDE_CURSOR)
            self._hidden_cursor = True

        self.current = min(current, self.total)
        progress = self.current / self.total if self.total > 0 else 0
        filled = int(self.width * progress)
        empty = self.width - filled

        # Build progress bar
        cfg = self.style_config

        if cfg.get("fill_color") is None:  # Gradient style
            bar_fill = "".join(
                self._get_gradient_char(i / self.width) for i in range(filled)
            )
            bar_fill += Colors.RESET
        else:
            bar_fill = f"{self.color}{cfg['fill'] * filled}{Colors.RESET}"

        bar_empty = f"{cfg['empty_color']}{cfg['empty'] * empty}{Colors.RESET}"
        bar = f"{cfg['left']}{bar_fill}{bar_empty}{cfg['right']}"

        # Build info string
        info_parts = []
        if self.show_percentage:
            info_parts.append(f"{progress * 100:5.1f}%")
        if self.show_count:
            info_parts.append(f"{self.current}/{self.total}")
        if self.show_eta and progress > 0:
            elapsed = time.time() - self._start_time
            eta = (elapsed / progress) * (1 - progress) if progress < 1 else 0
            info_parts.append(f"ETA: {self._format_time(eta)}")

        info = " ".join(info_parts)
        prefix = f"{self.prefix} " if self.prefix else ""

        # Build and print line
        line = f"\r{prefix}{bar} {info}"
        if message:
            line += f" {Colors.DIM}{message}{Colors.RESET}"

        sys.stdout.write(Colors.CLEAR_LINE + line)
        sys.stdout.flush()

    def finish(self, message: str = "Complete"):
        """Finish progress bar."""
        self.update(self.total)
        elapsed = time.time() - self._start_time
        sys.stdout.write(f" {Colors.BRIGHT_GREEN}✓{Colors.RESET} {message}")
        sys.stdout.write(f" {Colors.DIM}({self._format_time(elapsed)}){Colors.RESET}\n")
        sys.stdout.write(Colors.SHOW_CURSOR)
        sys.stdout.flush()

    def fail(self, message: str = "Failed"):
        """Mark progress as failed."""
        sys.stdout.write(f" {Colors.BRIGHT_RED}✗{Colors.RESET} {message}\n")
        sys.stdout.write(Colors.SHOW_CURSOR)
        sys.stdout.flush()


@dataclass
class TaskState:
    """State of a task in MultiProgress."""
    name: str
    total: int = 100
    current: int = 0
    status: str = "pending"  # pending, running, completed, failed
    color: str = Colors.BRIGHT_CYAN
    message: str = ""


class MultiProgress:
    """
    Multi-task progress display.

    Usage:
        tasks = [
            {'name': 'Nmap Scan', 'total': 100, 'color': Colors.BRIGHT_CYAN},
            {'name': 'Nikto', 'total': 100, 'color': Colors.BRIGHT_GREEN},
            {'name': 'Nuclei', 'total': 100, 'color': Colors.BRIGHT_YELLOW},
        ]
        mp = MultiProgress(tasks)
        mp.update(0, 50, status='running')
        mp.complete(0)
        mp.update(1, 75, status='running')
        mp.close()
    """

    STATUS_SYMBOLS = {
        "pending": ("○", Colors.DIM),
        "running": ("▸", Colors.BRIGHT_YELLOW),
        "completed": ("✓", Colors.BRIGHT_GREEN),
        "failed": ("✗", Colors.BRIGHT_RED),
        "skipped": ("⊘", Colors.DIM),
    }

    def __init__(
        self,
        tasks: List[Dict[str, Any]],
        bar_width: int = 30,
        name_width: int = 16,
    ):
        self.tasks = [
            TaskState(
                name=t.get("name", f"Task {i}"),
                total=t.get("total", 100),
                color=t.get("color", Colors.BRIGHT_CYAN),
            )
            for i, t in enumerate(tasks)
        ]
        self.bar_width = bar_width
        self.name_width = name_width
        self._displayed = False
        self._lock = threading.Lock()

    def _render_task(self, task: TaskState) -> str:
        """Render a single task line."""
        # Status symbol
        symbol, symbol_color = self.STATUS_SYMBOLS.get(
            task.status, self.STATUS_SYMBOLS["pending"]
        )

        # Name (padded)
        name = task.name[:self.name_width].ljust(self.name_width)

        # Progress bar
        progress = task.current / task.total if task.total > 0 else 0
        filled = int(self.bar_width * progress)
        empty = self.bar_width - filled

        bar = f"{task.color}{'▰' * filled}{Colors.RESET}{Colors.DIM}{'▱' * empty}{Colors.RESET}"

        # Percentage
        pct = f"{progress * 100:5.1f}%"

        return f"  {symbol_color}{symbol}{Colors.RESET} {name} {bar} {pct}"

    def _display(self):
        """Display all tasks."""
        with self._lock:
            if self._displayed:
                # Move cursor up to overwrite
                sys.stdout.write(f"\033[{len(self.tasks)}A")

            sys.stdout.write(Colors.HIDE_CURSOR)

            for task in self.tasks:
                line = self._render_task(task)
                sys.stdout.write(Colors.CLEAR_LINE + line + "\n")

            sys.stdout.flush()
            self._displayed = True

    def update(
        self,
        task_index: int,
        current: int,
        status: str = "running",
        message: str = "",
    ):
        """Update a task's progress."""
        if 0 <= task_index < len(self.tasks):
            self.tasks[task_index].current = current
            self.tasks[task_index].status = status
            self.tasks[task_index].message = message
            self._display()

    def complete(self, task_index: int, message: str = ""):
        """Mark a task as completed."""
        if 0 <= task_index < len(self.tasks):
            self.tasks[task_index].current = self.tasks[task_index].total
            self.tasks[task_index].status = "completed"
            self.tasks[task_index].message = message
            self._display()

    def fail(self, task_index: int, message: str = ""):
        """Mark a task as failed."""
        if 0 <= task_index < len(self.tasks):
            self.tasks[task_index].status = "failed"
            self.tasks[task_index].message = message
            self._display()

    def skip(self, task_index: int, message: str = ""):
        """Mark a task as skipped."""
        if 0 <= task_index < len(self.tasks):
            self.tasks[task_index].status = "skipped"
            self.tasks[task_index].message = message
            self._display()

    def close(self):
        """Close and cleanup."""
        sys.stdout.write(Colors.SHOW_CURSOR)
        sys.stdout.flush()


class TypeWriter:
    """
    Typewriter effect for text.

    Usage:
        TypeWriter.print("Initializing AIPTX...", speed=0.03)
    """

    @staticmethod
    def print(
        text: str,
        speed: float = 0.03,
        color: str = Colors.BRIGHT_GREEN,
        end: str = "\n",
    ):
        """Print text with typewriter effect."""
        sys.stdout.write(color)
        for char in text:
            sys.stdout.write(char)
            sys.stdout.flush()
            time.sleep(speed)
        sys.stdout.write(Colors.RESET + end)
        sys.stdout.flush()


class PulseText:
    """
    Pulsing/blinking text effect.

    Usage:
        pulse = PulseText("SCANNING", color=Colors.BRIGHT_RED)
        pulse.start()
        time.sleep(3)
        pulse.stop()
    """

    def __init__(self, text: str, color: str = Colors.BRIGHT_CYAN, speed: float = 0.5):
        self.text = text
        self.color = color
        self.speed = speed
        self._running = False
        self._thread: Optional[threading.Thread] = None

    def start(self):
        """Start pulsing."""
        self._running = True
        self._thread = threading.Thread(target=self._pulse, daemon=True)
        self._thread.start()

    def _pulse(self):
        """Pulse animation loop."""
        bright = True
        while self._running:
            if bright:
                sys.stdout.write(f"\r{self.color}{Colors.BOLD}{self.text}{Colors.RESET}")
            else:
                sys.stdout.write(f"\r{Colors.DIM}{self.text}{Colors.RESET}")
            sys.stdout.flush()
            bright = not bright
            time.sleep(self.speed)

    def stop(self, final_text: str = None):
        """Stop pulsing."""
        self._running = False
        if self._thread:
            self._thread.join(timeout=0.5)
        sys.stdout.write(f"\r{Colors.CLEAR_LINE}")
        if final_text:
            sys.stdout.write(final_text + "\n")
        sys.stdout.flush()


class MatrixRain:
    """
    Matrix-style rain effect (for fun loading screens).

    Usage:
        rain = MatrixRain(width=60, height=10)
        rain.start()
        time.sleep(3)
        rain.stop()
    """

    CHARS = "01アイウエオカキクケコサシスセソタチツテトナニヌネノハヒフヘホマミムメモヤユヨラリルレロワヲン"

    def __init__(
        self,
        width: int = 60,
        height: int = 10,
        color: str = Colors.BRIGHT_GREEN,
        speed: float = 0.05,
    ):
        self.width = width
        self.height = height
        self.color = color
        self.speed = speed
        self._running = False
        self._thread: Optional[threading.Thread] = None

    def start(self):
        """Start the rain."""
        self._running = True
        self._thread = threading.Thread(target=self._rain, daemon=True)
        self._thread.start()

    def _rain(self):
        """Rain animation loop."""
        columns = [0] * self.width
        sys.stdout.write(Colors.HIDE_CURSOR)

        while self._running:
            lines = []
            for y in range(self.height):
                line = ""
                for x in range(self.width):
                    if columns[x] == y:
                        line += f"{Colors.BRIGHT_WHITE}{random.choice(self.CHARS)}"
                    elif columns[x] > y and columns[x] - y < 5:
                        intensity = 255 - (columns[x] - y) * 50
                        line += f"\033[38;2;0;{intensity};0m{random.choice(self.CHARS)}"
                    else:
                        line += " "
                lines.append(line + Colors.RESET)

            # Update column positions
            for x in range(self.width):
                if random.random() > 0.95:
                    columns[x] = 0
                else:
                    columns[x] = (columns[x] + 1) % (self.height + 5)

            # Draw
            sys.stdout.write(f"\033[{self.height}A")
            for line in lines:
                sys.stdout.write(Colors.CLEAR_LINE + line + "\n")
            sys.stdout.flush()
            time.sleep(self.speed)

    def stop(self):
        """Stop the rain."""
        self._running = False
        if self._thread:
            self._thread.join(timeout=0.5)
        sys.stdout.write(Colors.SHOW_CURSOR)
        # Clear the rain area
        for _ in range(self.height):
            sys.stdout.write(Colors.CLEAR_LINE + "\n")
        sys.stdout.write(f"\033[{self.height}A")
        sys.stdout.flush()


# Convenience functions
def spin(message: str, style: str = "dots", color: str = Colors.BRIGHT_CYAN) -> Spinner:
    """Create and return a started spinner."""
    return Spinner(message, style=style, color=color).start()


def progress(total: int = 100, style: str = "cyber", **kwargs) -> ProgressBar:
    """Create a progress bar."""
    return ProgressBar(total=total, style=style, **kwargs)


# Demo function
def demo():
    """Demonstrate all animations."""
    print(f"\n{Colors.BOLD}AIPTX Animation Demo{Colors.RESET}\n")
    print("=" * 50)

    # Spinner demo
    print(f"\n{Colors.BRIGHT_YELLOW}1. Spinners:{Colors.RESET}")
    for style in ["dots", "cyber", "pulse", "snake"]:
        with Spinner(f"Testing {style} spinner...", style=style) as s:
            time.sleep(1)

    # Progress bar demo
    print(f"\n{Colors.BRIGHT_YELLOW}2. Progress Bars:{Colors.RESET}")
    for style in ["cyber", "hacker", "gradient", "neon"]:
        bar = ProgressBar(total=100, style=style, prefix=f"{style:10}")
        for i in range(101):
            bar.update(i)
            time.sleep(0.01)
        bar.finish()

    # Multi-progress demo
    print(f"\n{Colors.BRIGHT_YELLOW}3. Multi-Task Progress:{Colors.RESET}")
    tasks = [
        {"name": "Nmap Scan", "color": Colors.BRIGHT_CYAN},
        {"name": "Nikto", "color": Colors.BRIGHT_GREEN},
        {"name": "Nuclei", "color": Colors.BRIGHT_YELLOW},
    ]
    mp = MultiProgress(tasks)
    for i in range(101):
        for j, _ in enumerate(tasks):
            progress_val = min(i + j * 10, 100)
            status = "completed" if progress_val >= 100 else "running"
            mp.update(j, progress_val, status=status)
        time.sleep(0.02)
    mp.close()

    # Typewriter demo
    print(f"\n{Colors.BRIGHT_YELLOW}4. Typewriter:{Colors.RESET}")
    TypeWriter.print("Initializing AIPTX Beast Mode...", speed=0.02)

    print(f"\n{Colors.BRIGHT_GREEN}Demo complete!{Colors.RESET}\n")


if __name__ == "__main__":
    demo()



"""AIPT LLM Configuration"""

from __future__ import annotations
import os
from typing import Optional, List


class LLMConfig:
    """Configuration for LLM interactions"""

    def __init__(
        self,
        model_name: Optional[str] = None,
        enable_prompt_caching: bool = True,
        prompt_modules: Optional[List[str]] = None,
        timeout: Optional[int] = None,
    ):
        self.model_name = model_name or os.getenv("AIPT_LLM", "openai/gpt-4")

        if not self.model_name:
            raise ValueError("AIPT_LLM environment variable must be set and not empty")

        self.enable_prompt_caching = enable_prompt_caching
        self.prompt_modules = prompt_modules or []

        self.timeout = timeout or int(os.getenv("LLM_TIMEOUT", "300"))



"""
AIPTX Ollama Checkpoint Client
==============================

Specialized Ollama client for AI checkpoint operations.
Optimized for local LLM inference with streaming support.
"""

import asyncio
import json
import logging
from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Callable, Dict, List, Optional

import httpx

logger = logging.getLogger(__name__)


@dataclass
class OllamaConfig:
    """Configuration for Ollama client."""

    base_url: str = "http://localhost:11434"
    model: str = "mistral:7b"
    timeout: int = 300
    context_length: int = 8192
    temperature: float = 0.3
    num_predict: int = 2048
    top_p: float = 0.9
    repeat_penalty: float = 1.1


@dataclass
class OllamaResponse:
    """Response from Ollama API."""

    content: str
    model: str
    total_duration: int = 0  # nanoseconds
    prompt_eval_count: int = 0
    eval_count: int = 0
    done: bool = True

    @property
    def tokens_per_second(self) -> float:
        """Calculate tokens per second."""
        if self.total_duration > 0:
            return self.eval_count / (self.total_duration / 1e9)
        return 0.0


class OllamaCheckpointClient:
    """
    Ollama client optimized for checkpoint analysis.

    Features:
    - Streaming support for progress feedback
    - Automatic retry on failures
    - Context window management
    - Health checking

    Example:
        client = OllamaCheckpointClient(OllamaConfig(model="mistral:7b"))

        # Check health
        if await client.health_check():
            # Generate with streaming
            response = await client.analyze_with_streaming(
                prompt="Analyze these findings...",
                system_prompt="You are a security expert...",
                on_token=lambda t: print(t, end="", flush=True)
            )
    """

    # Model context window sizes (tokens)
    MODEL_CONTEXT_SIZES = {
        "llama2:7b": 4096,
        "llama2:13b": 4096,
        "llama2:70b": 4096,
        "llama3:8b": 8192,
        "llama3:70b": 8192,
        "llama3.2": 131072,
        "codellama:7b": 16384,
        "codellama:13b": 16384,
        "codellama:34b": 16384,
        "mistral:7b": 8192,
        "mistral:instruct": 8192,
        "mixtral:8x7b": 32768,
        "deepseek-coder:6.7b": 16384,
        "deepseek-coder:33b": 16384,
        "phi-3:3.8b": 4096,
        "phi-3:14b": 4096,
        "qwen2:7b": 32768,
        "qwen2.5:14b": 32768,
    }

    def __init__(self, config: Optional[OllamaConfig] = None):
        """Initialize the Ollama client."""
        self.config = config or OllamaConfig()
        self._available_models: List[str] = []

    async def health_check(self) -> bool:
        """
        Verify Ollama is running and model is available.

        Returns:
            True if Ollama is healthy and model is available
        """
        try:
            async with httpx.AsyncClient(timeout=5) as client:
                response = await client.get(f"{self.config.base_url}/api/tags")
                if response.status_code != 200:
                    return False

                data = response.json()
                self._available_models = [m.get("name", "") for m in data.get("models", [])]

                # Check if configured model is available
                model_base = self.config.model.split(":")[0]
                return any(model_base in m for m in self._available_models)

        except Exception as e:
            logger.warning(f"Ollama health check failed: {e}")
            return False

    async def get_available_models(self) -> List[str]:
        """Get list of available models."""
        if not self._available_models:
            await self.health_check()
        return self._available_models

    def get_context_size(self, model: Optional[str] = None) -> int:
        """Get context window size for a model."""
        model = model or self.config.model

        # Try exact match first
        if model in self.MODEL_CONTEXT_SIZES:
            return self.MODEL_CONTEXT_SIZES[model]

        # Try base model name
        base_model = model.split(":")[0]
        for key, value in self.MODEL_CONTEXT_SIZES.items():
            if key.startswith(base_model):
                return value

        # Default fallback
        return 4096

    async def generate(
        self,
        prompt: str,
        system_prompt: Optional[str] = None,
        model: Optional[str] = None,
    ) -> OllamaResponse:
        """
        Generate a response (non-streaming).

        Args:
            prompt: User prompt
            system_prompt: Optional system prompt
            model: Optional model override

        Returns:
            OllamaResponse object
        """
        messages = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})
        messages.append({"role": "user", "content": prompt})

        async with httpx.AsyncClient(timeout=self.config.timeout) as client:
            response = await client.post(
                f"{self.config.base_url}/api/chat",
                json={
                    "model": model or self.config.model,
                    "messages": messages,
                    "stream": False,
                    "options": {
                        "temperature": self.config.temperature,
                        "num_predict": self.config.num_predict,
                        "num_ctx": self.config.context_length,
                        "top_p": self.config.top_p,
                        "repeat_penalty": self.config.repeat_penalty,
                    },
                },
            )
            response.raise_for_status()
            data = response.json()

            return OllamaResponse(
                content=data.get("message", {}).get("content", ""),
                model=data.get("model", model or self.config.model),
                total_duration=data.get("total_duration", 0),
                prompt_eval_count=data.get("prompt_eval_count", 0),
                eval_count=data.get("eval_count", 0),
            )

    async def analyze_with_streaming(
        self,
        prompt: str,
        system_prompt: Optional[str] = None,
        model: Optional[str] = None,
        on_token: Optional[Callable[[str], None]] = None,
    ) -> str:
        """
        Stream response tokens for progress feedback.

        Args:
            prompt: User prompt with context
            system_prompt: Security-focused system instructions
            model: Optional model override
            on_token: Callback for each token (for UI progress)

        Returns:
            Complete response text
        """
        messages = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})
        messages.append({"role": "user", "content": prompt})

        full_response = []

        async with httpx.AsyncClient(timeout=self.config.timeout) as client:
            async with client.stream(
                "POST",
                f"{self.config.base_url}/api/chat",
                json={
                    "model": model or self.config.model,
                    "messages": messages,
                    "stream": True,
                    "options": {
                        "temperature": self.config.temperature,
                        "num_predict": self.config.num_predict,
                        "num_ctx": self.config.context_length,
                        "top_p": self.config.top_p,
                        "repeat_penalty": self.config.repeat_penalty,
                    },
                },
            ) as response:
                response.raise_for_status()

                async for line in response.aiter_lines():
                    if not line:
                        continue

                    try:
                        data = json.loads(line)
                        if "message" in data and "content" in data["message"]:
                            token = data["message"]["content"]
                            full_response.append(token)

                            if on_token:
                                on_token(token)

                        if data.get("done", False):
                            break

                    except json.JSONDecodeError:
                        continue

        return "".join(full_response)

    async def stream_tokens(
        self,
        prompt: str,
        system_prompt: Optional[str] = None,
        model: Optional[str] = None,
    ) -> AsyncIterator[str]:
        """
        Async iterator for streaming tokens.

        Args:
            prompt: User prompt
            system_prompt: Optional system prompt
            model: Optional model override

        Yields:
            Individual tokens
        """
        messages = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})
        messages.append({"role": "user", "content": prompt})

        async with httpx.AsyncClient(timeout=self.config.timeout) as client:
            async with client.stream(
                "POST",
                f"{self.config.base_url}/api/chat",
                json={
                    "model": model or self.config.model,
                    "messages": messages,
                    "stream": True,
                    "options": {
                        "temperature": self.config.temperature,
                        "num_predict": self.config.num_predict,
                        "num_ctx": self.config.context_length,
                    },
                },
            ) as response:
                response.raise_for_status()

                async for line in response.aiter_lines():
                    if not line:
                        continue

                    try:
                        data = json.loads(line)
                        if "message" in data and "content" in data["message"]:
                            yield data["message"]["content"]

                        if data.get("done", False):
                            break

                    except json.JSONDecodeError:
                        continue

    async def pull_model(
        self,
        model: str,
        progress_callback: Optional[Callable[[str, float], None]] = None,
    ) -> bool:
        """
        Pull a model from Ollama registry.

        Args:
            model: Model name to pull
            progress_callback: Optional callback(status, progress_percent)

        Returns:
            True if model pulled successfully
        """
        try:
            async with httpx.AsyncClient(timeout=3600) as client:  # 1 hour timeout
                async with client.stream(
                    "POST",
                    f"{self.config.base_url}/api/pull",
                    json={"name": model, "stream": True},
                ) as response:
                    response.raise_for_status()

                    async for line in response.aiter_lines():
                        if not line:
                            continue

                        try:
                            data = json.loads(line)
                            status = data.get("status", "")

                            if progress_callback:
                                total = data.get("total", 0)
                                completed = data.get("completed", 0)
                                if total > 0:
                                    percent = (completed / total) * 100
                                    progress_callback(status, percent)
                                else:
                                    progress_callback(status, 0)

                        except json.JSONDecodeError:
                            continue

            return True

        except Exception as e:
            logger.error(f"Failed to pull model {model}: {e}")
            return False

    def estimate_tokens(self, text: str) -> int:
        """
        Estimate token count for text.

        Uses approximate ratio of 4 characters per token.
        """
        return len(text) // 4



"""
AIPTX Thinking Tool
===================

Provides structured reasoning for complex decisions.
Helps agents think through multi-step problems systematically.

Integrated from Strix's thinking tool.

Usage:
    from aipt_v2.tools.thinking import think, analyze_options

    # Think through a decision
    result = think(
        reasoning="The target has an exposed API endpoint at /api/users...",
        decision="Test for IDOR vulnerability by modifying user IDs",
        confidence=0.85,
    )

    # Analyze multiple options
    result = analyze_options(
        context="Found potential SQL injection point",
        options=[
            "Use sqlmap for automated testing",
            "Manual testing with custom payloads",
            "Check for WAF before proceeding",
        ],
    )
"""

from .thinking_actions import (
    think,
    analyze_options,
    plan_attack,
    ThinkingResult,
)

__all__ = [
    "think",
    "analyze_options",
    "plan_attack",
    "ThinkingResult",
]



"""
Race Condition Test Patterns

Tests for TOCTOU (Time-of-Check to Time-of-Use) vulnerabilities,
double-spending, and concurrent operation issues.
"""

from aipt_v2.business_logic.patterns.base import (
    TestPattern,
    TestCase,
    PatternCategory,
    TestSeverity,
)


RACE_CONDITION_PATTERNS = [
    TestPattern(
        id="RACE-001",
        name="Double-Spend Attack",
        description="Test for double-spending by sending concurrent payment/withdrawal requests",
        category=PatternCategory.RACE_CONDITION,
        severity=TestSeverity.CRITICAL,
        cwe_ids=["CWE-362", "CWE-367"],
        owasp_category="Business Logic Errors",
        remediation="Implement proper transaction locking, use database transactions with appropriate isolation levels, implement idempotency keys",
        endpoint_patterns=[
            r"/payment", r"/transfer", r"/withdraw", r"/checkout",
            r"/purchase", r"/order", r"/transaction"
        ],
        applicable_to=["e-commerce", "banking", "payment"],
        test_cases=[
            TestCase(
                name="Concurrent Withdrawal",
                description="Send multiple withdrawal requests simultaneously",
                method="POST",
                endpoint_pattern=r"/(withdraw|transfer|payment)",
                body_template={"amount": 100, "account": "{{account_id}}"},
                concurrent_requests=10,
                success_indicators=["success", "completed", "approved"],
                failure_indicators=["insufficient", "balance", "failed"]
            ),
            TestCase(
                name="Concurrent Order Placement",
                description="Place multiple orders for limited stock item",
                method="POST",
                endpoint_pattern=r"/(order|checkout|purchase)",
                body_template={"item_id": "{{item_id}}", "quantity": 1},
                concurrent_requests=5,
                success_indicators=["order_id", "confirmed"],
                failure_indicators=["out_of_stock", "unavailable"]
            )
        ]
    ),

    TestPattern(
        id="RACE-002",
        name="Coupon Race Condition",
        description="Apply same single-use coupon multiple times via concurrent requests",
        category=PatternCategory.RACE_CONDITION,
        severity=TestSeverity.HIGH,
        cwe_ids=["CWE-362"],
        owasp_category="Business Logic Errors",
        remediation="Implement coupon redemption with database locks, use atomic operations",
        endpoint_patterns=[
            r"/coupon", r"/promo", r"/discount", r"/voucher", r"/redeem"
        ],
        applicable_to=["e-commerce", "marketing"],
        test_cases=[
            TestCase(
                name="Concurrent Coupon Application",
                description="Apply single-use coupon to multiple carts simultaneously",
                method="POST",
                endpoint_pattern=r"/(apply.*coupon|redeem|promo)",
                body_template={"code": "{{coupon_code}}", "cart_id": "{{cart_id}}"},
                concurrent_requests=5,
                success_indicators=["applied", "discount", "success"],
                failure_indicators=["already_used", "invalid", "expired"]
            )
        ]
    ),

    TestPattern(
        id="RACE-003",
        name="Bonus/Reward Race",
        description="Claim signup bonus or reward multiple times",
        category=PatternCategory.RACE_CONDITION,
        severity=TestSeverity.HIGH,
        cwe_ids=["CWE-362", "CWE-367"],
        owasp_category="Business Logic Errors",
        remediation="Use atomic claim operations, implement proper state management",
        endpoint_patterns=[
            r"/bonus", r"/reward", r"/claim", r"/redeem", r"/spin"
        ],
        applicable_to=["gaming", "loyalty", "marketing"],
        test_cases=[
            TestCase(
                name="Concurrent Bonus Claim",
                description="Claim one-time bonus through concurrent requests",
                method="POST",
                endpoint_pattern=r"/(claim|redeem|bonus)",
                body_template={"reward_id": "{{reward_id}}"},
                concurrent_requests=10,
                success_indicators=["claimed", "credited", "success"],
                failure_indicators=["already_claimed", "not_eligible"]
            )
        ]
    ),

    TestPattern(
        id="RACE-004",
        name="Vote/Like Race Condition",
        description="Submit multiple votes or likes bypassing single-vote restriction",
        category=PatternCategory.RACE_CONDITION,
        severity=TestSeverity.MEDIUM,
        cwe_ids=["CWE-362"],
        owasp_category="Business Logic Errors",
        remediation="Use unique constraints and atomic operations for vote tracking",
        endpoint_patterns=[
            r"/vote", r"/like", r"/upvote", r"/rate"
        ],
        applicable_to=["social", "voting", "content"],
        test_cases=[
            TestCase(
                name="Concurrent Votes",
                description="Submit multiple votes simultaneously",
                method="POST",
                endpoint_pattern=r"/(vote|like|upvote)",
                body_template={"target_id": "{{target_id}}", "direction": "up"},
                concurrent_requests=5,
                success_indicators=["voted", "success"],
                failure_indicators=["already_voted", "duplicate"]
            )
        ]
    ),

    TestPattern(
        id="RACE-005",
        name="Password Reset Race",
        description="Exploit race condition in password reset token validation",
        category=PatternCategory.RACE_CONDITION,
        severity=TestSeverity.HIGH,
        cwe_ids=["CWE-362", "CWE-640"],
        owasp_category="Authentication Flaws",
        remediation="Invalidate token immediately upon first use, use atomic operations",
        endpoint_patterns=[
            r"/reset", r"/password", r"/forgot"
        ],
        applicable_to=["authentication"],
        test_cases=[
            TestCase(
                name="Concurrent Reset Token Use",
                description="Use same reset token in concurrent password changes",
                method="POST",
                endpoint_pattern=r"/(reset.*password|password.*reset)",
                body_template={
                    "token": "{{reset_token}}",
                    "new_password": "NewPass123!"
                },
                concurrent_requests=3,
                success_indicators=["changed", "updated", "success"],
                failure_indicators=["invalid_token", "expired", "already_used"]
            )
        ]
    ),
]



"""
AIPTX SPA Scanner - Single-Page Application Security Testing

Provides comprehensive SPA security testing:
- Browser-based crawling with Playwright
- DOM-based XSS detection
- Client-side routing analysis
- State management inspection
- API request interception
- Source map analysis
"""

from aipt_v2.scanners.spa.scanner import (
    SPAScanner,
    SPAScanConfig,
    SPAScanResult,
    DOMXSSFinding,
)

__all__ = [
    "SPAScanner",
    "SPAScanConfig",
    "SPAScanResult",
    "DOMXSSFinding",
]



"""
AIPTX Beast Mode - Exploitation Engine
=======================================

LLM-powered multi-step exploitation with adaptive payload mutation.

Components:
- chain_builder: Plan multi-step attack chains (SQLi→RCE, SSRF→Exfil)
- chain_executor: Execute planned chains with fallback
- chain_templates: Pre-defined attack chain patterns
- payload_engine: WAF-aware payload mutation
- waf_detector: WAF fingerprinting and bypass selection
- encoders: Encoding/obfuscation techniques
"""

from __future__ import annotations

from aipt_v2.exploitation.chain_builder import (
    ExploitChainBuilder,
    ChainStep,
    AttackChain,
)
from aipt_v2.exploitation.chain_executor import ChainExecutor
from aipt_v2.exploitation.chain_templates import (
    CHAIN_TEMPLATES,
    get_chain_template,
    ChainTemplate,
)
from aipt_v2.exploitation.payload_engine import PayloadEngine
from aipt_v2.exploitation.encoders import (
    Encoder,
    url_encode,
    base64_encode,
    hex_encode,
    unicode_encode,
    double_url_encode,
    chain_encode,
)
from aipt_v2.exploitation.waf_detector import WAFDetector
from aipt_v2.exploitation.mutations import (
    SQLiMutator,
    XSSMutator,
    CMDMutator,
)

__all__ = [
    # Chain Builder
    "ExploitChainBuilder",
    "ChainStep",
    "AttackChain",
    # Chain Executor
    "ChainExecutor",
    # Templates
    "CHAIN_TEMPLATES",
    "get_chain_template",
    "ChainTemplate",
    # Payload Engine
    "PayloadEngine",
    # Encoders
    "Encoder",
    "url_encode",
    "base64_encode",
    "hex_encode",
    "unicode_encode",
    "double_url_encode",
    "chain_encode",
    # WAF Detection
    "WAFDetector",
    # Mutations
    "SQLiMutator",
    "XSSMutator",
    "CMDMutator",
]



"""
OWASP Top 10 2021 Mapping

Provides detailed OWASP Top 10 2021 category definitions
and CWE-to-OWASP mapping.

Usage:
    from aipt_v2.compliance import OWASPMapper, get_owasp_category

    category = get_owasp_category("CWE-79")  # Returns A03
"""

from dataclasses import dataclass, field
from typing import List, Dict, Optional


@dataclass
class OWASPCategory:
    """OWASP Top 10 category definition."""
    id: str
    name: str
    description: str
    cwes: List[str]
    risk_factors: Dict[str, str]
    prevention: List[str]


# OWASP Top 10 2021 definitions
OWASP_TOP_10 = {
    "A01": OWASPCategory(
        id="A01:2021",
        name="Broken Access Control",
        description="Access control enforces policy such that users cannot act outside "
                   "of their intended permissions. Failures typically lead to unauthorized "
                   "information disclosure, modification, or destruction of all data or "
                   "performing a business function outside the user's limits.",
        cwes=["CWE-22", "CWE-23", "CWE-35", "CWE-59", "CWE-200", "CWE-201", "CWE-219",
              "CWE-264", "CWE-275", "CWE-276", "CWE-284", "CWE-285", "CWE-352", "CWE-359",
              "CWE-377", "CWE-402", "CWE-425", "CWE-441", "CWE-497", "CWE-538", "CWE-540",
              "CWE-548", "CWE-552", "CWE-566", "CWE-601", "CWE-639", "CWE-651", "CWE-668",
              "CWE-706", "CWE-862", "CWE-863", "CWE-913", "CWE-922", "CWE-1275"],
        risk_factors={
            "threat_agents": "Anyone with network access",
            "exploitability": "Average",
            "prevalence": "Widespread",
            "detectability": "Average",
            "impact": "Severe"
        },
        prevention=[
            "Except for public resources, deny by default",
            "Implement access control mechanisms once and re-use them",
            "Enforce record ownership, rather than accepting that the user can create, read, update, or delete any record",
            "Disable web server directory listing and ensure file metadata and backup files are not present",
            "Log access control failures, alert admins when appropriate",
            "Rate limit API and controller access to minimize harm from automated attack tooling",
            "Invalidate JWT tokens on the server after logout"
        ]
    ),

    "A02": OWASPCategory(
        id="A02:2021",
        name="Cryptographic Failures",
        description="Failures related to cryptography (or lack thereof) which often lead "
                   "to exposure of sensitive data. This includes exposure of sensitive data "
                   "that requires protection, such as passwords, credit card numbers, health "
                   "records, personal information.",
        cwes=["CWE-261", "CWE-296", "CWE-310", "CWE-319", "CWE-320", "CWE-321", "CWE-322",
              "CWE-323", "CWE-324", "CWE-325", "CWE-326", "CWE-327", "CWE-328", "CWE-329",
              "CWE-330", "CWE-331", "CWE-335", "CWE-336", "CWE-337", "CWE-338", "CWE-340",
              "CWE-347", "CWE-523", "CWE-720", "CWE-757", "CWE-759", "CWE-760", "CWE-780",
              "CWE-818", "CWE-916"],
        risk_factors={
            "threat_agents": "Attackers with access to data in transit/at rest",
            "exploitability": "Average",
            "prevalence": "Widespread",
            "detectability": "Average",
            "impact": "Severe"
        },
        prevention=[
            "Classify data processed, stored, or transmitted by an application",
            "Don't store sensitive data unnecessarily. Discard it as soon as possible",
            "Make sure to encrypt all sensitive data at rest",
            "Ensure up-to-date and strong standard algorithms, protocols, and keys are in place",
            "Encrypt all data in transit with secure protocols such as TLS",
            "Disable caching for responses that contain sensitive data",
            "Do not use legacy protocols such as FTP and SMTP for transporting sensitive data"
        ]
    ),

    "A03": OWASPCategory(
        id="A03:2021",
        name="Injection",
        description="User-supplied data is not validated, filtered, or sanitized by the "
                   "application. Dynamic queries or non-parameterized calls without "
                   "context-aware escaping are used directly in the interpreter.",
        cwes=["CWE-20", "CWE-74", "CWE-75", "CWE-77", "CWE-78", "CWE-79", "CWE-80",
              "CWE-83", "CWE-87", "CWE-88", "CWE-89", "CWE-90", "CWE-91", "CWE-93",
              "CWE-94", "CWE-95", "CWE-96", "CWE-97", "CWE-98", "CWE-99", "CWE-113",
              "CWE-116", "CWE-138", "CWE-184", "CWE-470", "CWE-471", "CWE-564", "CWE-610",
              "CWE-643", "CWE-644", "CWE-652", "CWE-917"],
        risk_factors={
            "threat_agents": "Anyone who can send untrusted data",
            "exploitability": "Easy",
            "prevalence": "Common",
            "detectability": "Easy",
            "impact": "Severe"
        },
        prevention=[
            "Use a safe API which avoids using the interpreter entirely",
            "Use positive server-side input validation",
            "Use LIMIT and other SQL controls to prevent mass disclosure",
            "Use parameterized queries and stored procedures",
            "Escape special characters using the specific escape syntax"
        ]
    ),

    "A04": OWASPCategory(
        id="A04:2021",
        name="Insecure Design",
        description="Insecure design is a broad category representing different weaknesses, "
                   "expressed as 'missing or ineffective control design.' This category "
                   "focuses on risks related to design and architectural flaws.",
        cwes=["CWE-73", "CWE-183", "CWE-209", "CWE-213", "CWE-235", "CWE-256", "CWE-257",
              "CWE-266", "CWE-269", "CWE-280", "CWE-311", "CWE-312", "CWE-313", "CWE-316",
              "CWE-419", "CWE-430", "CWE-434", "CWE-444", "CWE-451", "CWE-472", "CWE-501",
              "CWE-522", "CWE-525", "CWE-539", "CWE-579", "CWE-598", "CWE-602", "CWE-642",
              "CWE-646", "CWE-650", "CWE-653", "CWE-656", "CWE-657", "CWE-799", "CWE-807",
              "CWE-840", "CWE-841", "CWE-927", "CWE-1021", "CWE-1173"],
        risk_factors={
            "threat_agents": "Varies based on design flaw",
            "exploitability": "Average",
            "prevalence": "Common",
            "detectability": "Difficult",
            "impact": "Moderate to Severe"
        },
        prevention=[
            "Establish and use a secure development lifecycle",
            "Use threat modeling for critical authentication and access control",
            "Integrate security language and controls into user stories",
            "Write unit and integration tests to validate security controls",
            "Tier application and network layers for critical applications"
        ]
    ),

    "A05": OWASPCategory(
        id="A05:2021",
        name="Security Misconfiguration",
        description="The application might be vulnerable if missing appropriate security "
                   "hardening or having improperly configured permissions on cloud services. "
                   "Default configurations, incomplete or ad hoc configurations.",
        cwes=["CWE-2", "CWE-11", "CWE-13", "CWE-15", "CWE-16", "CWE-260", "CWE-315",
              "CWE-520", "CWE-526", "CWE-537", "CWE-541", "CWE-547", "CWE-611", "CWE-614",
              "CWE-756", "CWE-776", "CWE-942", "CWE-1004", "CWE-1032", "CWE-1174"],
        risk_factors={
            "threat_agents": "Attackers with system access",
            "exploitability": "Easy",
            "prevalence": "Widespread",
            "detectability": "Easy",
            "impact": "Moderate"
        },
        prevention=[
            "A repeatable hardening process for fast and easy deployment",
            "A minimal platform without unnecessary features and components",
            "A task to review and update configurations as part of patch management",
            "A segmented application architecture with effective separation",
            "Sending security directives to clients, e.g., Security Headers"
        ]
    ),

    "A06": OWASPCategory(
        id="A06:2021",
        name="Vulnerable and Outdated Components",
        description="Components run with the same privileges as the application. "
                   "If a vulnerable component is exploited, such an attack can "
                   "facilitate serious data loss or server takeover.",
        cwes=["CWE-937", "CWE-1035", "CWE-1104"],
        risk_factors={
            "threat_agents": "Attackers with vulnerability knowledge",
            "exploitability": "Average",
            "prevalence": "Widespread",
            "detectability": "Difficult",
            "impact": "Moderate to Severe"
        },
        prevention=[
            "Remove unused dependencies, unnecessary features, components",
            "Continuously inventory component versions (client and server-side)",
            "Monitor sources like CVE and NVD for vulnerabilities",
            "Only obtain components from official sources over secure links",
            "Monitor for libraries and components that are unmaintained"
        ]
    ),

    "A07": OWASPCategory(
        id="A07:2021",
        name="Identification and Authentication Failures",
        description="Confirmation of the user's identity, authentication, and session "
                   "management is critical to protect against authentication-related attacks.",
        cwes=["CWE-255", "CWE-259", "CWE-287", "CWE-288", "CWE-290", "CWE-294", "CWE-295",
              "CWE-297", "CWE-300", "CWE-302", "CWE-304", "CWE-306", "CWE-307", "CWE-346",
              "CWE-384", "CWE-521", "CWE-613", "CWE-620", "CWE-640", "CWE-798", "CWE-940",
              "CWE-1216"],
        risk_factors={
            "threat_agents": "Anyone attempting to impersonate users",
            "exploitability": "Average",
            "prevalence": "Common",
            "detectability": "Average",
            "impact": "Severe"
        },
        prevention=[
            "Implement multi-factor authentication",
            "Do not ship or deploy with any default credentials",
            "Implement weak-password checks against top 10,000 worst passwords",
            "Use a server-side session manager that generates random session IDs",
            "Limit or increasingly delay failed login attempts"
        ]
    ),

    "A08": OWASPCategory(
        id="A08:2021",
        name="Software and Data Integrity Failures",
        description="Code and infrastructure that does not protect against integrity "
                   "violations. This includes using plugins, libraries, or modules from "
                   "untrusted sources, repositories, and CDNs.",
        cwes=["CWE-345", "CWE-353", "CWE-426", "CWE-494", "CWE-502", "CWE-565", "CWE-784",
              "CWE-829", "CWE-830", "CWE-915"],
        risk_factors={
            "threat_agents": "Supply chain or CI/CD attackers",
            "exploitability": "Average",
            "prevalence": "Common",
            "detectability": "Difficult",
            "impact": "Severe"
        },
        prevention=[
            "Use digital signatures to verify software or data is from expected source",
            "Ensure libraries and dependencies are consuming trusted repositories",
            "Use a software supply chain security tool like OWASP Dependency-Check",
            "Ensure your CI/CD pipeline has proper segregation and access control",
            "Do not send unsigned or unencrypted serialized data to untrusted clients"
        ]
    ),

    "A09": OWASPCategory(
        id="A09:2021",
        name="Security Logging and Monitoring Failures",
        description="This category helps detect, escalate, and respond to active breaches. "
                   "Without logging and monitoring, breaches cannot be detected.",
        cwes=["CWE-117", "CWE-223", "CWE-532", "CWE-778"],
        risk_factors={
            "threat_agents": "Attackers relying on lack of monitoring",
            "exploitability": "Average",
            "prevalence": "Widespread",
            "detectability": "Difficult",
            "impact": "Moderate"
        },
        prevention=[
            "Ensure all login, access control, and server-side input validation failures are logged",
            "Ensure logs are generated in a format easily consumed by log management solutions",
            "Ensure log data is encoded correctly to prevent injections",
            "Establish effective monitoring and alerting",
            "Establish an incident response and recovery plan"
        ]
    ),

    "A10": OWASPCategory(
        id="A10:2021",
        name="Server-Side Request Forgery",
        description="SSRF flaws occur whenever a web application is fetching a remote "
                   "resource without validating the user-supplied URL. It allows an "
                   "attacker to coerce the application to send a crafted request to "
                   "an unexpected destination.",
        cwes=["CWE-918"],
        risk_factors={
            "threat_agents": "Attackers with access to URL input",
            "exploitability": "Average",
            "prevalence": "Common",
            "detectability": "Average",
            "impact": "Moderate to Severe"
        },
        prevention=[
            "Sanitize and validate all client-supplied input data",
            "Enforce the URL schema, port, and destination with a positive allow list",
            "Do not send raw responses to clients",
            "Disable HTTP redirections",
            "Use network-level firewall policies to block all but essential traffic"
        ]
    )
}


class OWASPMapper:
    """OWASP Top 10 specific mapper."""

    def __init__(self):
        self.categories = OWASP_TOP_10

    def get_category(self, cwe_id: str) -> Optional[OWASPCategory]:
        """Get OWASP category for a CWE."""
        cwe_id = cwe_id.upper()
        if not cwe_id.startswith("CWE-"):
            cwe_id = f"CWE-{cwe_id}"

        for cat_id, category in self.categories.items():
            if cwe_id in category.cwes:
                return category

        return None

    def get_category_by_id(self, category_id: str) -> Optional[OWASPCategory]:
        """Get OWASP category by ID (A01-A10)."""
        return self.categories.get(category_id.upper())

    def get_all_cwes_for_category(self, category_id: str) -> List[str]:
        """Get all CWEs mapped to a category."""
        category = self.categories.get(category_id.upper())
        return category.cwes if category else []


def get_owasp_category(cwe_id: str) -> Optional[str]:
    """
    Get OWASP category ID for a CWE.

    Args:
        cwe_id: CWE identifier

    Returns:
        OWASP category ID (A01-A10) or None
    """
    mapper = OWASPMapper()
    category = mapper.get_category(cwe_id)
    return category.id if category else None



"""
AIPTX Interactive Shell
=======================

Provides an interactive shell environment for running security tools
with enhanced features like command history, output logging, and
real-time display.

Features:
- Interactive command execution
- Tab completion for tool names
- Command history with readline
- Real-time output streaming
- Session logging
- Tool discovery and help
- Environment management
- Cross-platform support (Windows, Linux, macOS)

Usage:
    aiptx shell                    # Start interactive shell
    aiptx shell --log session.log  # With logging
"""

import asyncio
import os
import shlex
import shutil
import signal
import subprocess
import sys
import platform
import threading
from datetime import datetime
from pathlib import Path
from queue import Queue, Empty
from typing import Dict, List, Optional, Tuple

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.syntax import Syntax
from rich.live import Live
from rich.text import Text
from rich import box


# Platform detection
IS_WINDOWS = platform.system() == "Windows"

# Conditionally import Unix-only modules
if not IS_WINDOWS:
    import pty
    import select
    import termios
    import tty

# Try to import readline (available on Unix, may need pyreadline3 on Windows)
try:
    import readline
    HAS_READLINE = True
except ImportError:
    HAS_READLINE = False
    # On Windows, try pyreadline3
    if IS_WINDOWS:
        try:
            import pyreadline3 as readline
            HAS_READLINE = True
        except ImportError:
            pass


console = Console()


# Command history file
HISTORY_FILE = Path.home() / ".aiptx" / "shell_history"


class ToolCompleter:
    """Tab completion for tool names and commands."""

    def __init__(self, tools: List[str]):
        self.tools = sorted(tools)
        self.commands = [
            "help", "tools", "exit", "quit", "clear", "history",
            "env", "cd", "pwd", "run", "scan", "export", "log"
        ]
        self.all_completions = self.commands + self.tools
        self.matches = []

    def complete(self, text: str, state: int) -> Optional[str]:
        """Readline completion function."""
        if state == 0:
            if text:
                self.matches = [
                    s for s in self.all_completions
                    if s.startswith(text)
                ]
            else:
                self.matches = self.all_completions[:]

        try:
            return self.matches[state]
        except IndexError:
            return None


class InteractiveShell:
    """
    Interactive shell for running security tools.

    Provides a REPL-like environment with:
    - Tool execution with real-time output
    - Command history and tab completion
    - Session logging
    - Environment management
    - Cross-platform compatibility
    """

    def __init__(
        self,
        log_file: Optional[Path] = None,
        working_dir: Optional[Path] = None,
    ):
        """
        Initialize the interactive shell.

        Args:
            log_file: Optional file to log session output
            working_dir: Working directory (defaults to current)
        """
        self.log_file = log_file
        self.working_dir = working_dir or Path.cwd()
        self.running = False
        self.env = os.environ.copy()
        self.history: List[str] = []
        self._tools: Dict[str, str] = {}
        self._load_tools()
        self._setup_readline()

    def _load_tools(self):
        """Load available tools from the installer."""
        try:
            from aipt_v2.local_tool_installer import TOOLS
            self._tools = {
                name: tool.description
                for name, tool in TOOLS.items()
            }
        except ImportError:
            self._tools = {}

    def _setup_readline(self):
        """Configure readline for history and completion."""
        if not HAS_READLINE:
            return

        # Load history
        HISTORY_FILE.parent.mkdir(parents=True, exist_ok=True)
        if HISTORY_FILE.exists():
            try:
                readline.read_history_file(str(HISTORY_FILE))
            except Exception:
                pass

        # Set up completion
        tool_names = list(self._tools.keys())
        completer = ToolCompleter(tool_names)
        readline.set_completer(completer.complete)
        readline.parse_and_bind("tab: complete")
        readline.set_completer_delims(" \t\n;")

        # Limit history size
        readline.set_history_length(1000)

    def _save_history(self):
        """Save command history to file."""
        if not HAS_READLINE:
            return
        try:
            readline.write_history_file(str(HISTORY_FILE))
        except Exception:
            pass

    def _log(self, message: str):
        """Log message to file if logging is enabled."""
        if self.log_file:
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            with open(self.log_file, "a") as f:
                f.write(f"[{timestamp}] {message}\n")

    def print_banner(self):
        """Print welcome banner with hacker aesthetic."""
        from rich.align import Align
        from rich.panel import Panel
        from rich.text import Text
        import socket
        import os
        from datetime import datetime

        # Hacker colors
        NEON_GREEN = "#00ff41"
        DARK_GREEN = "#008f11"
        CYBER_BLUE = "#00d4ff"
        BLOOD_RED = "#ff0040"
        GHOST_WHITE = "#c0c0c0"

        term_width = console.size.width

        # Cyber banner
        banner = f"""
[bold {NEON_GREEN}]╔══════════════════════════════════════════════════════════════════════════╗[/]
[bold {NEON_GREEN}]║[/]  [bold {CYBER_BLUE}]█▀█ █ █▀█ ▀█▀ ▀▄▀   [/][bold white]INTERACTIVE SHELL[/]                                [bold {NEON_GREEN}]║[/]
[bold {NEON_GREEN}]║[/]  [{BLOOD_RED}]█▀█ █ █▀▀  █  █ █   [/][dim {GHOST_WHITE}]Tool Execution Environment[/]                       [bold {NEON_GREEN}]║[/]
[bold {NEON_GREEN}]╚══════════════════════════════════════════════════════════════════════════╝[/]"""

        console.clear()
        console.print(Align.center(banner))

        # System info
        try:
            hostname = socket.gethostname()
            username = os.getenv("USER") or os.getenv("USERNAME") or "operator"
        except Exception:
            hostname = "localhost"
            username = "operator"

        sys_info = Text()
        sys_info.append("  ┌─", style=f"dim {DARK_GREEN}")
        sys_info.append(f" {username}@{hostname}", style=f"bold {NEON_GREEN}")
        sys_info.append(" │ ", style=f"dim {DARK_GREEN}")
        sys_info.append(f"cwd: {self.working_dir}", style=f"dim {GHOST_WHITE}")
        sys_info.append(" ─┐", style=f"dim {DARK_GREEN}")

        console.print()
        console.print(Align.center(sys_info))
        console.print()
        console.print(f"[dim {DARK_GREEN}]" + "─" * term_width + f"[/]")
        console.print(Align.center(f"[dim]Type [bold {NEON_GREEN}]help[/] for commands • [bold {NEON_GREEN}]tools[/] to list available arsenal[/dim]"))

        if IS_WINDOWS:
            console.print(Align.center(f"[dim {BLOOD_RED}][!] Windows detected - some features limited[/]"))

        console.print()

    def print_help(self):
        """Print help information with hacker aesthetic."""
        from rich.table import Table
        from rich import box

        NEON_GREEN = "#00ff41"
        DARK_GREEN = "#008f11"
        CYBER_BLUE = "#00d4ff"
        BLOOD_RED = "#ff0040"
        GHOST_WHITE = "#c0c0c0"

        console.print()
        console.print(f"[bold {CYBER_BLUE}]╔══ SHELL COMMAND REFERENCE ══╗[/]")
        console.print()

        # Built-in commands table
        cmd_table = Table(
            show_header=True,
            header_style=f"bold {NEON_GREEN}",
            box=box.SIMPLE_HEAD,
            border_style=DARK_GREEN,
            expand=True,
        )
        cmd_table.add_column("COMMAND", style=f"bold {NEON_GREEN}")
        cmd_table.add_column("ACTION", style=GHOST_WHITE)

        cmd_table.add_row("help", "Display this reference")
        cmd_table.add_row("tools [category]", "List available security tools")
        cmd_table.add_row("clear", "Purge terminal buffer")
        cmd_table.add_row("history", "Display command history")
        cmd_table.add_row("env", "Show environment variables")
        cmd_table.add_row("env SET KEY=VAL", "Set environment variable")
        cmd_table.add_row("cd <path>", "Change working directory")
        cmd_table.add_row("pwd", "Print working directory")
        cmd_table.add_row("log <file>", "Start session logging")
        cmd_table.add_row("exit", "Terminate shell session")

        console.print(cmd_table)
        console.print()

        # Usage examples
        console.print(f"[bold {BLOOD_RED}]⚔ TOOL EXECUTION[/]")
        console.print(f"[dim {GHOST_WHITE}]  Execute tools directly by name:[/]")
        console.print(f"    [{NEON_GREEN}]nmap -sV -sC target.com[/]")
        console.print(f"    [{NEON_GREEN}]nuclei -u https://target.com -t cves/[/]")
        console.print(f"    [{NEON_GREEN}]sqlmap -u \"http://target.com?id=1\" --batch[/]")
        console.print()

        # Hotkeys
        console.print(f"[bold {CYBER_BLUE}]⌨ HOTKEYS[/]")
        console.print(f"  [dim]TAB[/]    → Auto-completion")
        console.print(f"  [dim]↑/↓[/]    → Navigate history")
        console.print(f"  [dim]Ctrl+C[/] → Kill current process")
        console.print(f"  [dim]Ctrl+D[/] → Exit shell")
        console.print()

    def list_tools(self, category: Optional[str] = None):
        """List available tools with hacker aesthetic."""
        try:
            from aipt_v2.local_tool_installer import TOOLS, ToolCategory

            NEON_GREEN = "#00ff41"
            DARK_GREEN = "#008f11"
            CYBER_BLUE = "#00d4ff"
            BLOOD_RED = "#ff0040"
            GHOST_WHITE = "#c0c0c0"

            title = f"[bold {BLOOD_RED}]⚔ ARSENAL[/] - Security Tools"
            if category:
                title += f" [{category.upper()}]"

            table = Table(
                title=title,
                box=box.HEAVY_EDGE,
                border_style=DARK_GREEN,
                header_style=f"bold {CYBER_BLUE}",
                expand=True,
            )
            table.add_column("TOOL", style=f"bold {NEON_GREEN}")
            table.add_column("CATEGORY", style=f"dim {GHOST_WHITE}")
            table.add_column("STATUS", justify="center")
            table.add_column("DESCRIPTION", style=GHOST_WHITE)

            for name, tool in sorted(TOOLS.items()):
                if category and tool.category.value != category:
                    continue

                is_installed = shutil.which(name) is not None
                status = f"[bold {NEON_GREEN}]●[/]" if is_installed else f"[dim {BLOOD_RED}]○[/]"

                desc = tool.description[:45] + "..." if len(tool.description) > 45 else tool.description

                table.add_row(name, tool.category.value, status, desc)

            console.print()
            console.print(table)
            console.print()

            if not category:
                console.print("[dim]Filter by category: tools <category>[/dim]")
                categories = sorted(set(t.category.value for t in TOOLS.values()))
                console.print(f"[dim]Categories: {', '.join(categories)}[/dim]")

        except ImportError:
            console.print("[yellow]Tool catalog not available[/yellow]")

    def show_history(self, count: int = 20):
        """Show command history."""
        if not HAS_READLINE:
            console.print("[yellow]Readline not available - history limited[/yellow]")
            # Show from our own history list
            recent = self.history[-count:] if len(self.history) > count else self.history
            console.print(f"\n[bold]Last {len(recent)} commands:[/bold]")
            for i, cmd in enumerate(recent, 1):
                console.print(f"  {i:4d}  {cmd}")
            console.print()
            return

        history_len = readline.get_current_history_length()
        start = max(1, history_len - count + 1)

        console.print(f"\n[bold]Last {min(count, history_len)} commands:[/bold]")
        for i in range(start, history_len + 1):
            cmd = readline.get_history_item(i)
            console.print(f"  {i:4d}  {cmd}")
        console.print()

    def show_env(self):
        """Show relevant environment variables."""
        relevant_prefixes = ["AIPT", "PATH", "HOME", "USER", "SHELL", "GOPATH", "GOBIN"]
        table = Table(title="Environment Variables", box=box.ROUNDED)
        table.add_column("Variable", style="cyan")
        table.add_column("Value", style="green", overflow="fold")

        for key, value in sorted(self.env.items()):
            if any(key.startswith(p) for p in relevant_prefixes):
                # Truncate long values
                display_val = value[:80] + "..." if len(value) > 80 else value
                table.add_row(key, display_val)

        console.print()
        console.print(table)
        console.print()

    def set_env(self, key: str, value: str):
        """Set an environment variable."""
        self.env[key] = value
        os.environ[key] = value
        console.print(f"[green]Set {key}={value}[/green]")

    def change_dir(self, path: str):
        """Change working directory."""
        try:
            new_path = Path(path).expanduser().resolve()
            if new_path.is_dir():
                os.chdir(new_path)
                self.working_dir = new_path
                console.print(f"[green]Changed to: {new_path}[/green]")
            else:
                console.print(f"[red]Not a directory: {path}[/red]")
        except Exception as e:
            console.print(f"[red]Error: {e}[/red]")

    async def run_command(self, command: str) -> int:
        """
        Run a command with real-time output.

        Uses PTY on Unix for proper terminal handling, allowing interactive
        tools to work correctly. Uses subprocess with threading on Windows.

        Args:
            command: Command to execute

        Returns:
            Exit code of the command
        """
        self._log(f"$ {command}")

        # Parse the command
        try:
            parts = shlex.split(command)
            if not parts:
                return 0
            program = parts[0]
        except ValueError as e:
            console.print(f"[red]Parse error: {e}[/red]")
            return 1

        # Check if tool exists
        if not shutil.which(program):
            console.print(f"[yellow]Command not found: {program}[/yellow]")

            # Check if it's a known tool that's not installed
            if program in self._tools:
                console.print(f"[dim]This tool is not installed. Run: aiptx tools install -t {program}[/dim]")
            return 127

        console.print()

        if IS_WINDOWS:
            return await self._run_command_windows(command)
        else:
            return await self._run_command_unix(command)

    async def _run_command_windows(self, command: str) -> int:
        """
        Run command on Windows using subprocess with threading.

        Args:
            command: Command to execute

        Returns:
            Exit code
        """
        try:
            process = subprocess.Popen(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                env=self.env,
                cwd=str(self.working_dir),
                creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
            )

            output_lines = []
            queue = Queue()
            stop_event = threading.Event()

            def reader_thread():
                try:
                    while not stop_event.is_set():
                        line = process.stdout.readline()
                        if line:
                            queue.put(line)
                        elif process.poll() is not None:
                            break
                except Exception:
                    pass
                finally:
                    queue.put(None)

            thread = threading.Thread(target=reader_thread, daemon=True)
            thread.start()

            try:
                while True:
                    try:
                        line = queue.get(timeout=0.1)
                        if line is None:
                            break
                        text = line.decode("utf-8", errors="replace")
                        sys.stdout.write(text)
                        sys.stdout.flush()
                        output_lines.append(text.rstrip())
                        self._log(text.rstrip())
                    except Empty:
                        if process.poll() is not None:
                            # Drain remaining
                            while True:
                                try:
                                    line = queue.get_nowait()
                                    if line is None:
                                        break
                                    text = line.decode("utf-8", errors="replace")
                                    sys.stdout.write(text)
                                    sys.stdout.flush()
                                    output_lines.append(text.rstrip())
                                except Empty:
                                    break
                            break

            except KeyboardInterrupt:
                # Send CTRL_BREAK_EVENT to process group
                try:
                    subprocess.run(
                        ["taskkill", "/F", "/T", "/PID", str(process.pid)],
                        capture_output=True,
                        timeout=5
                    )
                except Exception:
                    process.kill()
                console.print("\n[yellow]Interrupted[/yellow]")

            finally:
                stop_event.set()
                thread.join(timeout=1)
                process.wait()

            return process.returncode or 0

        except Exception as e:
            console.print(f"[red]Error running command: {e}[/red]")
            return 1

    async def _run_command_unix(self, command: str) -> int:
        """
        Run command on Unix using PTY for full terminal support.

        Args:
            command: Command to execute

        Returns:
            Exit code
        """
        try:
            # Create pseudo-terminal
            master_fd, slave_fd = pty.openpty()

            # Start process
            process = subprocess.Popen(
                command,
                shell=True,
                stdin=slave_fd,
                stdout=slave_fd,
                stderr=slave_fd,
                env=self.env,
                cwd=str(self.working_dir),
                preexec_fn=os.setsid,
            )

            os.close(slave_fd)

            # Read output in real-time
            output_lines = []
            try:
                while True:
                    ready, _, _ = select.select([master_fd], [], [], 0.1)
                    if ready:
                        try:
                            data = os.read(master_fd, 1024)
                            if not data:
                                break
                            text = data.decode("utf-8", errors="replace")
                            sys.stdout.write(text)
                            sys.stdout.flush()
                            output_lines.append(text)
                            self._log(text.rstrip())
                        except OSError:
                            break

                    # Check if process finished
                    if process.poll() is not None:
                        # Read any remaining output
                        try:
                            while True:
                                ready, _, _ = select.select([master_fd], [], [], 0.1)
                                if not ready:
                                    break
                                data = os.read(master_fd, 1024)
                                if not data:
                                    break
                                text = data.decode("utf-8", errors="replace")
                                sys.stdout.write(text)
                                sys.stdout.flush()
                                output_lines.append(text)
                        except OSError:
                            pass
                        break

            except KeyboardInterrupt:
                # Send SIGINT to process group
                os.killpg(os.getpgid(process.pid), signal.SIGINT)
                console.print("\n[yellow]Interrupted[/yellow]")

            finally:
                os.close(master_fd)
                process.wait()

            return process.returncode or 0

        except Exception as e:
            console.print(f"[red]Error running command: {e}[/red]")
            return 1

    def process_builtin(self, command: str) -> Tuple[bool, int]:
        """
        Process built-in commands.

        Returns:
            Tuple of (is_builtin, exit_code)
        """
        parts = command.strip().split(maxsplit=1)
        if not parts:
            return True, 0

        cmd = parts[0].lower()
        args = parts[1] if len(parts) > 1 else ""

        if cmd in ("exit", "quit"):
            self.running = False
            return True, 0

        elif cmd == "help":
            self.print_help()
            return True, 0

        elif cmd == "tools":
            self.list_tools(args if args else None)
            return True, 0

        elif cmd == "clear":
            console.clear()
            return True, 0

        elif cmd == "history":
            count = int(args) if args.isdigit() else 20
            self.show_history(count)
            return True, 0

        elif cmd == "env":
            if args.startswith("SET ") or args.startswith("set "):
                # Set environment variable
                var_part = args[4:].strip()
                if "=" in var_part:
                    key, value = var_part.split("=", 1)
                    self.set_env(key.strip(), value.strip())
                else:
                    console.print("[yellow]Usage: env SET KEY=VALUE[/yellow]")
            else:
                self.show_env()
            return True, 0

        elif cmd == "cd":
            if args:
                self.change_dir(args)
            else:
                self.change_dir(str(Path.home()))
            return True, 0

        elif cmd == "pwd":
            console.print(str(self.working_dir))
            return True, 0

        elif cmd == "log":
            if args:
                self.log_file = Path(args).expanduser()
                console.print(f"[green]Logging to: {self.log_file}[/green]")
            else:
                if self.log_file:
                    console.print(f"[dim]Currently logging to: {self.log_file}[/dim]")
                else:
                    console.print("[dim]Logging disabled. Use: log <filename>[/dim]")
            return True, 0

        return False, 0

    async def run(self):
        """Run the interactive shell."""
        self.running = True
        self.print_banner()

        while self.running:
            try:
                # Build prompt
                cwd = str(self.working_dir)
                home = str(Path.home())
                if cwd.startswith(home):
                    cwd = "~" + cwd[len(home):]

                prompt = f"\n[bold green]aiptx[/bold green]:[bold blue]{cwd}[/bold blue]$ "
                console.print(prompt, end="")

                # Read input
                try:
                    command = input().strip()
                except EOFError:
                    console.print()
                    break

                if not command:
                    continue

                # Add to history
                self.history.append(command)

                # Check for built-in commands
                is_builtin, exit_code = self.process_builtin(command)
                if is_builtin:
                    continue

                # Run external command
                exit_code = await self.run_command(command)

                if exit_code != 0:
                    console.print(f"[dim]Exit code: {exit_code}[/dim]")

            except KeyboardInterrupt:
                console.print("\n[dim]Use 'exit' or Ctrl+D to quit[/dim]")
                continue

        # Save history on exit
        self._save_history()
        console.print("\n[dim]Goodbye![/dim]")


async def start_interactive_shell(
    log_file: Optional[str] = None,
    working_dir: Optional[str] = None,
) -> int:
    """
    Start the interactive shell.

    Args:
        log_file: Optional log file path
        working_dir: Optional working directory

    Returns:
        Exit code
    """
    shell = InteractiveShell(
        log_file=Path(log_file) if log_file else None,
        working_dir=Path(working_dir) if working_dir else None,
    )
    await shell.run()
    return 0


def main():
    """CLI entry point for interactive shell."""
    import argparse

    parser = argparse.ArgumentParser(description="AIPTX Interactive Shell")
    parser.add_argument(
        "--log", "-l",
        help="Log session to file"
    )
    parser.add_argument(
        "--dir", "-d",
        help="Working directory"
    )

    args = parser.parse_args()

    asyncio.run(start_interactive_shell(
        log_file=args.log,
        working_dir=args.dir,
    ))


if __name__ == "__main__":
    main()
